/* Desc: Postgres module
INCLUDES=-I/usr/local/pgsql/include
EXTRA_LIBS=-L/usr/local/pgsql/lib -lpq -lcrypt

Both postgres and javascript defines int8,int16, etc.
To compile: add #if 0 to c.h in postgres


TODO
* support user/passwords
* support cursors: get_row()


*/


#include <stdlib.h>

#include <mod_js_config.h>

#undef int8
#undef int16
#undef int32
#undef uint8
#undef uint16
#undef uint32
#undef float32
#undef float64
#include "libpq-fe.h"
#include "jsmodules.h"



typedef struct {
	PGconn *connection;
}Postgres_info;

typedef struct {
	PGresult *result;
	int field_number;
}Postgres_field;

/*******************
result from connection...
*/

static void finalize_postgres_result(JSContext *cx, JSObject *obj)
{
	PGresult *result;
	result=JS_GetPrivate(cx,obj);
/*~~~ delete the result fields */
	PQclear(result);
}

static void finalize_postgres_field(JSContext *cx, JSObject *obj)
{
	Postgres_field *field;
	field=JS_GetPrivate(cx,obj);
	JS_free(cx,field);
}

static JSBool get_postgres_result_field(JSContext *cx,JSObject *obj,jsval id, jsval *val)
{
	if(JSVAL_IS_INT(id)) {
		Postgres_field *field_info;

		field_info=JS_GetPrivate(cx,obj);
		*val=STRING_TO_JSVAL(JS_NewStringCopyZ(cx,PQgetvalue(field_info->result,
			JSVAL_TO_INT(id),field_info->field_number)));
		return JS_TRUE;
	}
	else {
		JS_ReportError(cx,"not a field number: %s\n",
		 JS_GetStringBytes(JS_ValueToString(cx,id)));
	}
	return JS_FALSE;
}



/* a field in the postgres results */
static JSClass postgres_field_class = {
	"Postgres_field",
	JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  get_postgres_result_field,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   finalize_postgres_field
};

static JSClass postgres_result_class = {
	"Postgres_result",
	JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   finalize_postgres_result
};

/* make a new field */
static JSObject *make_pg_field(JSContext *cx,PGresult *this,int field_number)
{
	JSObject *field_obj;
	Postgres_field *field_info;

	field_obj=JS_NewObject(cx,&postgres_field_class,NULL,NULL);
	field_info=JS_malloc(cx,sizeof(*field_info));
	field_info->field_number=field_number;
	field_info->result=this;
	JS_SetPrivate(cx,field_obj,field_info);

	return field_obj;
}


static JSObject *make_pg_result(JSContext *cx,PGresult *this)
{
	JSObject *result;
	int nfields;
	int c;

	if((result=JS_NewObject(cx,&postgres_result_class,NULL,NULL))==NULL) {
		return NULL;
	}
	JS_SetPrivate(cx,result,this);

	/* go through all the fields and set the tuples */
	if(PQresultStatus(this)==PGRES_TUPLES_OK) {
		nfields=PQnfields(this);
		for(c=0; c<nfields; c++) {
			char *name;
			JSObject *field;
			jsval field_jsval;

			name=PQfname(this,c);
			field=make_pg_field(cx,this,c);
			field_jsval=OBJECT_TO_JSVAL(field);
			JS_SetProperty(cx,result,name,&field_jsval);
		}
	}

	return result;
}


/*******************
connection...
*/


static void finalize_postgres(JSContext *cx, JSObject *obj)
{
	Postgres_info *postgres_info;
	if((postgres_info=JS_GetPrivate(cx,obj))!=NULL) {
/* ~~~ permanent cache for connections */
		PQfinish(postgres_info->connection);
		JS_free(cx,postgres_info);
	}
}


static JSClass postgres_class = {
	NULL,
	JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   finalize_postgres
};

/* connect to a db */
static JSBool
Postgres(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	char *host=NULL,*port=NULL,*options=NULL,*tty=NULL,*dbname=NULL;
	char *user=NULL,*password=NULL;
	Postgres_info *postgres_info;
	PGconn *connection;

	JS_GC(cx);
	if(argc>=1) {
		dbname=JS_GetStringBytes(JS_ValueToString(cx,argv[0]));
	}
	if(argc>=2) {
		host=JS_GetStringBytes(JS_ValueToString(cx,argv[1]));
	}
	if(argc>=3) {
		user=JS_GetStringBytes(JS_ValueToString(cx,argv[2]));
	}
	if(argc>=4) {
		password=JS_GetStringBytes(JS_ValueToString(cx,argv[3]));
	}
	if(argc>=5) {
		options=JS_GetStringBytes(JS_ValueToString(cx,argv[4]));
	}

/* ~~~ permanent cache for connections */
	connection=PQsetdb(host,port,options,tty,dbname);
	if(PQstatus(connection)!=CONNECTION_OK) {
		*rval=JSVAL_NULL;
		JS_ReportError(cx,"Cannot connect to db\n");
		PQfinish(connection);
		return JS_TRUE;
	}
	postgres_info=JS_malloc(cx,sizeof(*postgres_info));
	postgres_info->connection=connection;

	JS_SetPrivate(cx,obj,postgres_info);
	*rval=OBJECT_TO_JSVAL(obj);
	return JS_TRUE;
}

static JSBool
postgres_exec(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	int err;
	const char *query;
	PGresult *result;
	Postgres_info *postgres_info;

	postgres_info=JS_GetPrivate(cx,obj);
	if(argc>0) {
		query=JS_GetStringBytes(JSVAL_TO_STRING(argv[0]));
	}

	result=PQexec(postgres_info->connection,query);
	err=PQresultStatus(result);
	if(err!=PGRES_COMMAND_OK && err!=PGRES_TUPLES_OK
	 && err!=PGRES_COPY_IN && err!=PGRES_COPY_OUT) {
		JS_ReportError(cx,"Postgres error: %s\nquery: %s",
		 PQerrorMessage(postgres_info->connection),query);
		*rval=JSVAL_NULL;
		return JS_TRUE;
	}

	*rval=OBJECT_TO_JSVAL(make_pg_result(cx,result));
	return JS_TRUE;
}

static JSBool
postgres_error(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	Postgres_info *postgres_info;
	const char *error;

	postgres_info=JS_GetPrivate(cx,obj);
	error=PQerrorMessage(postgres_info->connection);
	*rval=STRING_TO_JSVAL(JS_NewStringCopyZ(cx,error));
	return JS_TRUE;
}

static JSBool
result_ntuples(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	PGresult *result;
	if(!JSVAL_IS_OBJECT(argv[0])) {
		JS_ReportError(cx,"ntuples: needs result object\n");
		return JS_FALSE;
	}
	result=JS_GetPrivate(cx,JSVAL_TO_OBJECT(argv[0]));
	*rval=INT_TO_JSVAL(PQntuples(result));
	return JS_TRUE;
}

static JSFunctionSpec postgres_static_functions[]={
	{"ntuples",	result_ntuples,	1},
	{NULL}
};


static JSFunctionSpec postgres_functions[]={
	{"exec",postgres_exec,1},
	{"errorMessage",postgres_error,0},
	{NULL}
};


static JSObject *
InitPostgresModule(JSContext *cx, JSObject *obj)
{
	JSObject *proto;

	proto = JS_InitClass(cx, obj, NULL, &postgres_class, Postgres, 0,
		NULL, postgres_functions, NULL, postgres_static_functions);
	if (!proto) return NULL;

	return proto;
}


Javascript_module Postgres_javascript_module={
	&postgres_class,
	&InitPostgresModule,
	NULL,
};



