/* desc: Gdbm module
*/

#include "mod_js_config.h"
#include <gdbm.h>
#include <errno.h>


/*************/

#define JSGDBM_UNICODE 0x10000		/* whether to store into gdbm as unicode or bytes */
#define JSGDBM_MASK 0xffff
typedef struct Gdbm_info {
	GDBM_FILE handle;
	int flags;
}Gdbm_info;

static JSClass gdbm_class;

/*************/



static JSBool
jsgdbm_enumerate(JSContext *cx, JSObject *obj,JSIterateOp enum_op,jsval *statep,jsid *idp)
{
	datum key,*prev_key;
	Gdbm_info *gdbm_info;

	if((gdbm_info=JS_GetPrivate(cx,obj))==NULL) { 
		*statep=JSVAL_NULL;
		return JS_TRUE; 
	}
	if(enum_op==JSENUMERATE_INIT) {
		*statep=JSVAL_NULL;
		if(idp!=NULL) { *idp=JSVAL_ZERO; }
	}
	else if(enum_op==JSENUMERATE_NEXT) {
		if(JSVAL_IS_NULL(*statep)) {
			/* get first key */
			key=gdbm_firstkey(gdbm_info->handle);
			prev_key=JS_malloc(cx,sizeof(*prev_key));
			*prev_key=key;
			*statep=PRIVATE_TO_JSVAL(prev_key);
			*idp=(jsid)js_Atomize(cx,key.dptr,key.dsize,0);
		}
		else {
			prev_key=JSVAL_TO_PRIVATE(*statep);
			key=gdbm_nextkey(gdbm_info->handle,*prev_key);
			free(prev_key->dptr);
			JS_free(cx,prev_key);

			if(key.dptr!=NULL) {
				prev_key=JS_malloc(cx,sizeof(*prev_key));
				*prev_key=key;
				*statep=PRIVATE_TO_JSVAL(prev_key);
				*idp=(jsid)js_Atomize(cx,key.dptr,key.dsize,0);
			}
			else {
				*statep=JSVAL_NULL;
			}
		}
	}
	else if(enum_op==JSENUMERATE_DESTROY) {
		if(!JSVAL_IS_NULL(*statep)) {
			prev_key=JSVAL_TO_PRIVATE(*statep);
			free(prev_key->dptr);
			JS_free(cx,prev_key);
		}
	}
	return JS_TRUE;
}

/*************/

static JSBool
jsgdbm_get_property(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
	Gdbm_info *gdbm_info;
	datum key,value;
	JSString *key_str,*value_str;

	if((gdbm_info=JS_GetPrivate(cx,obj))==NULL) { return JS_TRUE; }
	if((key_str=JS_ValueToString(cx,id))==NULL) {
		return JS_FALSE;
	}
	key.dptr=JS_GetStringBytes(key_str);
	key.dsize=JS_GetStringLength(key_str);

	value=gdbm_fetch(gdbm_info->handle,key);
	if(value.dptr==NULL)  { *vp=JSVAL_NULL; return JS_TRUE; }

	if((value_str=JS_NewStringCopyN(cx,value.dptr,value.dsize))==NULL) { return JS_FALSE; }
	free(value.dptr);

	*vp=STRING_TO_JSVAL(value_str);
	return JS_TRUE;
}

static JSBool
jsgdbm_set_property(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
	Gdbm_info *gdbm_info;
	datum key,value;
	JSString *key_str,*value_str;

	if((gdbm_info=JS_GetPrivate(cx,obj))==NULL) { return JS_FALSE; }
	if((key_str=JS_ValueToString(cx,id))==NULL
	 || (value_str=JS_ValueToString(cx,*vp))==NULL
	) {
		return JS_FALSE;
	}

	key.dsize=JS_GetStringLength(key_str);
	value.dsize=JS_GetStringLength(value_str);

	if(gdbm_info->flags&JSGDBM_UNICODE) {
		key.dptr=(char *)JS_GetStringChars(key_str);
		value.dptr=(char *)JS_GetStringChars(value_str);
	}
	else {
		key.dptr=JS_GetStringBytes(key_str);
		value.dptr=JS_GetStringBytes(value_str);
	}
	if(gdbm_store(gdbm_info->handle,key,value,GDBM_REPLACE)!=0) {
		JS_ReportError(cx,"Problem storing data into gdbm: err: %s\n",
		 gdbm_strerror(gdbm_errno));
	}
	*vp=JSVAL_TRUE;
	return JS_TRUE;
}

static void
jsgdbm_close(JSContext *cx, JSObject *obj)
{
	Gdbm_info *gdbm_info;

	if((gdbm_info=JS_GetPrivate(cx,obj))==NULL) { return; }
	gdbm_close(gdbm_info->handle);
}

static JSBool
Gdbm(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	GDBM_FILE handle;
	JSString *filename_str;
	char *filename;
	jsint mode;
	Gdbm_info *gdbm_info;
	int flags;
	jsint block_size,read_write;
	
	/* collect garbage to close off previous opened Gdbm */
	JS_GC(cx);
	if(argc<3) {
		JS_ReportError(cx,"Not enough arguments to for Gdbm()\n");
		return JS_FALSE;
	}

	if((filename_str=JS_ValueToString(cx,argv[0]))==NULL
	 || (filename=JS_GetStringBytes(filename_str))==NULL
	) {
		JS_ReportError(cx,"Gdbm: Cannot get filename\n");
		return JS_FALSE;
	}
	if(JS_ValueToInt32(cx,argv[1],&block_size)!=JS_TRUE
	 || JS_ValueToInt32(cx,argv[2],&read_write)!=JS_TRUE
	) {
		JS_ReportError(cx,"Gdbm: Cannot get block size or read/write flags\n");
		return JS_FALSE;
	}
	flags=read_write;
	read_write&=JSGDBM_MASK;
	mode=0640;
	if(argc>=4) {
		JS_ValueToInt32(cx,argv[3],&mode);
	}

	if((gdbm_info=JS_malloc(cx,sizeof(gdbm_info)))==NULL
	 || JS_SetPrivate(cx,obj,gdbm_info)==JS_FALSE
	) {
		return JS_FALSE;
	}
	gdbm_info->flags=flags;

	if((handle=gdbm_open(filename,block_size,read_write,mode,NULL))==NULL) {
		JS_ReportError(cx,"Cannot open gdbm err:%s, %s\n",gdbm_strerror(gdbm_errno),sys_errlist[errno]);
		return JS_FALSE;
	}
	gdbm_info->handle=handle;
	*rval=OBJECT_TO_JSVAL(obj);
	return JS_TRUE;
}

/*************/

static JSClass gdbm_class = {
	NULL,
	JSCLASS_HAS_PRIVATE|JSCLASS_NEW_ENUMERATE,
	JS_PropertyStub,  JS_PropertyStub, jsgdbm_get_property, jsgdbm_set_property,
	(JSEnumerateOp)jsgdbm_enumerate, JS_ResolveStub,   JS_ConvertStub,   jsgdbm_close
};


static JSConstDoubleSpec gdbm_flags[] = {
	{GDBM_READER,"READER"},
	{GDBM_WRITER,"WRITER"},
	{GDBM_WRCREAT,"WRCREAT"},
	{GDBM_NEWDB,"NEWDB"},
	{GDBM_FAST,"FAST"},
	{JSGDBM_UNICODE,"UNICODE"},
	{0,NULL},
};



static JSObject *
InitGdbmModule(JSContext *cx, JSObject *obj)
{
	JSObject *proto,*ctor;

	proto = JS_InitClass(cx, obj, NULL, &gdbm_class, Gdbm, 0,
		NULL, NULL, NULL,NULL);
	if (!proto || (ctor=JS_GetConstructor(cx,proto))==NULL) return NULL;
	if(JS_DefineConstDoubles(cx,ctor,gdbm_flags)!=JS_TRUE) {
		return NULL;
	}
	return proto;
}


Javascript_module Gdbm_javascript_module={
	&gdbm_class,
	&InitGdbmModule,
	NULL,
};

