/* desc: Mysql module
*/

#include <mod_js_config.h>
#include <mysql.h>

static void mysql_result_finalize(JSContext *cx, JSObject *obj);
static void mysql_finalize(JSContext *cx, JSObject *obj);

static JSClass mysql_class = {
	NULL,
	JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   mysql_finalize
};

static JSClass mysql_result_class = {
	"Mysql_result",
	JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   mysql_result_finalize
};

/****************/



/****************/

static void
mysql_result_finalize(JSContext *cx, JSObject *obj)
{
	MYSQL_RES *res;
	if((res=JS_GetPrivate(cx,obj))==NULL)  { return; }
	mysql_free_result(res);
}

/* do a sql query */
static JSBool
mysql_exec(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MYSQL *mysql;
	MYSQL_RES *res;
	MYSQL_ROW row;
	JSString *query_str;
	char *query;
	JSObject **fields;
	int nfields;
	int c;
	int row_num;
	JSObject *result;

	if((mysql=JS_GetPrivate(cx,obj))==NULL)  { return JS_FALSE; }
	query_str=JS_ValueToString(cx,argv[0]);
/*~~~ option here to not put everything into properties and use fetch_rows for cursors */

	query=JS_GetStringBytes(query_str);
	if(mysql_query(mysql,query)) {
		JS_ReportError(cx,"Mysql Cannot do query: %s (err:%s)\n",query,mysql_error(mysql));
		*rval=JSVAL_NULL;
		return JS_TRUE;
	}
	if(mysql->status!=MYSQL_STATUS_GET_RESULT) {
		/* not a query that returns a result so just exit */
		*rval=JSVAL_TRUE;
		return JS_TRUE;
	}

	if(!(res=mysql_store_result(mysql))) {
		JS_ReportError(cx,"Cannot store results: %s\n",mysql_error(mysql));
		*rval=JSVAL_NULL;
		return JS_TRUE;
	}

	/* put the results into an object */
	if((result=JS_NewObject(cx,&mysql_result_class,NULL,NULL))==NULL) { return JS_FALSE; }
	JS_SetPrivate(cx,result,res);

	nfields=mysql_num_fields(res);
	fields=JS_malloc(cx,sizeof(*fields)*nfields);
	for(c=0; c<nfields; c++) {
		JSObject *f;
		jsval v;
		MYSQL_FIELD *field;

		fields[c]=f=JS_NewArrayObject(cx,0,NULL);
/*~~~ use table name too? */
		v=OBJECT_TO_JSVAL(f);
		field=mysql_fetch_field(res);
		JS_SetProperty(cx,result,field->name,&v);
	}
	row_num=0;
	while((row=mysql_fetch_row(res))) {
		for(c=0; c<nfields; c++) {
			jsval v;
			v=STRING_TO_JSVAL(JS_NewStringCopyZ(cx,row[c]));
			JS_SetElement(cx,fields[c],row_num,&v);
		}
		row_num++;
	}
	JS_free(cx,fields);

	*rval=OBJECT_TO_JSVAL(result);
	return JS_TRUE;
}

static JSBool
mysql_ntuples(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MYSQL_RES *res;
	if(!JSVAL_IS_OBJECT(argv[0])) {
		JS_ReportError(cx,"ntuples: Must point to a result object\n");
		return JS_FALSE;
	}
	if((res=JS_GetPrivate(cx,JSVAL_TO_OBJECT(argv[0])))==NULL) {
		JS_ReportError(cx,"ntuples: Is an object but not a result object\n");
		return JS_FALSE;
	}
	*rval=INT_TO_JSVAL(mysql_num_rows(res));
	return JS_TRUE;
}

/****************/

static JSFunctionSpec mysql_methods[] = {
	{"exec",		mysql_exec,		1},
	{NULL,NULL,0}
};

static JSFunctionSpec mysql_static_methods[] = {
	{"ntuples",		mysql_ntuples,		1},
	{NULL,NULL,0}
};


/****************/

static void
mysql_finalize(JSContext *cx, JSObject *obj)
{
	MYSQL *mysql;
	if((mysql=JS_GetPrivate(cx,obj))==NULL)  { return; }
	mysql_close(mysql);
	JS_free(cx,mysql);
}

static JSBool
Mysql(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MYSQL *mysql;
	char *dbase,*host=NULL,*user=NULL,*password=NULL;
	JSString *dbase_str,*host_name,*password_str,*user_str;

	JS_GC(cx);	/* close old Mysql connection first */
	if((mysql=JS_malloc(cx,sizeof(*mysql)))==NULL) { return JS_FALSE; }
	if(argc<1) {
		JS_ReportError(cx,"Not enough arguments to Mysql\n");
		return JS_FALSE;
	}
	if((dbase_str=JS_ValueToString(cx,argv[0]))==NULL
	 || (dbase=JS_GetStringBytes(dbase_str))==NULL
	) {
		JS_ReportError(cx,"Cannot get dbase name from argv\n");
		return JS_FALSE;
	}
	
	if(argc>=2) {
		if(JSVAL_IS_NULL(argv[1])) {
			host=NULL;
		}
		else if((host_name=JS_ValueToString(cx,argv[1]))==NULL
		 || (host=JS_GetStringBytes(host_name))==NULL
		) {
			JS_ReportError(cx,"Cannot get hostname from argv\n");
			return JS_FALSE;
		}
	}
	if(argc>=3) {
		if(JSVAL_IS_NULL(argv[2])) {
			user=NULL;
		}
		if((user_str=JS_ValueToString(cx,argv[2]))==NULL
		 || (user=JS_GetStringBytes(user_str))==NULL
		) {
			JS_ReportError(cx,"Cannot get username from argv\n");
			return JS_FALSE;
		}
	}
	if(argc>=4) {
		if(JSVAL_IS_NULL(argv[3])) {
			password=NULL;
		}
		else if((password_str=JS_ValueToString(cx,argv[3]))==NULL
		 || (password=JS_GetStringBytes(password_str))==NULL
		) {
			JS_ReportError(cx,"Cannot get password from argv\n");
			return JS_FALSE;
		}
	}

	if(!mysql_connect(mysql,host,user,password)) {
		JS_ReportError(cx,"Cannot connect to mysql:%s\n",mysql_error(mysql));
		return JS_FALSE;
	}
	if(mysql_select_db(mysql,dbase)) {
		JS_ReportError(cx,"Mysql: Cannot select dbase:%s :%s\n",dbase,mysql_error(mysql));
		return JS_FALSE;
	}
	if(JS_SetPrivate(cx,obj,mysql)!=JS_TRUE) { return JS_FALSE; }

	*rval=OBJECT_TO_JSVAL(obj);
	return JS_TRUE;
}




static JSObject *
InitMysqlModule(JSContext *cx, JSObject *obj)
{
	JSObject *proto,*result_proto;

	proto = JS_InitClass(cx, obj, NULL, &mysql_class,Mysql, 0,
		NULL, mysql_methods, NULL, mysql_static_methods);
	if (!proto) return NULL;
	result_proto = JS_InitClass(cx, obj, NULL, &mysql_result_class, NULL, 0,
		NULL, NULL, NULL, NULL);
	if (!result_proto) return NULL;

	return proto;
}


Javascript_module Mysql_javascript_module={
	&mysql_class,
	&InitMysqlModule,
	NULL,
};

